from airflow import DAG

from airflow.operators.postgres_operator import PostgresOperator
from airflow.operators.bash_operator import BashOperator
from airflow.sensors.sql_sensor import SqlSensor
from airflow.hooks.base_hook import BaseHook
from airflow.configuration import conf
from airflow.models import Variable

from ewah.constants import EWAHConstants as EC

from datetime import datetime, timedelta
import os

def dbt_dags_factory(
    dwh_engine,
    dwh_conn_id,
    project_name,
    dbt_schema_name,
    airflow_conn_id,
    dag_base_name='DBT_run',
    analytics_reader=None, # list of users of DWH who are read-only
    schedule_interval=timedelta(hours=1),
    start_date=datetime(2019,1,1),
    default_args=None,
    folder=None,
    models=None,
    exclude=None,
):

    if analytics_reader:
        for statement in ('insert', 'update', 'delete',
            'drop', 'create', 'select', ';', 'grant'):
            for reader in analytics_reader:
                if (statement in reader.lower()):
                    raise Exception('Error! The analytics reader {0} ' \
                        + 'is invalid.'.format(reader))

        #analytics_reader = analytics_reader.split(',')
        analytics_reader_sql = f'\nGRANT USAGE ON SCHEMA "{dbt_schema_name}"'
        analytics_reader_sql += ' TO "{0}";'
        analytics_reader_sql += f'''
        \nGRANT SELECT ON ALL TABLES IN SCHEMA "{dbt_schema_name}"''' + ' TO "{0}";'
        analytics_reader_sql = ''.join([
            analytics_reader_sql.format(i) for i in analytics_reader
        ])

    if models and not (type(models) == str):
        models = ' --models ' + ' '.join(models)
    else:
        models = ''

    if exclude and not (type(exclude) == str):
        exclude = ' --exclude ' + ' '.join(exclude)
    else:
        exclude = ''

    flags = models + exclude

    dag = DAG(
        dag_base_name,
        catchup=False,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        default_args=default_args,
    )

    dag_full_refresh = DAG(
        dag_base_name + '_full_refresh',
        catchup=False,
        max_active_runs=1,
        schedule_interval=None,
        start_date=start_date,
        default_args=default_args,
    )

    folder = folder or (os.environ.get('AIRFLOW_HOME') or conf.get('core', 'airflow_home')).replace(
        'airflow_home/airflow',
        'dbt_home',
    )

    bash_command = '''
    cd {1}
    source env/bin/activate
    cd {2}
    dbt {0}
    '''.format(
        '{0}',
        folder,
        project_name,
    )

    sensor_sql = """
        SELECT
            CASE WHEN COUNT(*) = 0 THEN 1 ELSE 0 END -- only run if exatly equal to 0
        FROM public.dag_run
        WHERE dag_id IN ('{0}', '{1}')
        and state = 'running'
        and not (run_id = '{2}')
    """.format(
        dag._dag_id,
        dag_full_refresh._dag_id,
        '{{ run_id }}',
    )

    # refactor?! not coupled to values in profiles.yml!
    if dwh_engine == EC.DWH_ENGINE_POSTGRES:
        conn = BaseHook.get_connection(dwh_conn_id)
        env = {
            'DBT_DWH_HOST': str(conn.host),
            'DBT_DWH_USER': str(conn.login),
            'DBT_DWH_PASS': str(conn.password),
            'DBT_DWH_PORT': str(conn.port),
            'DBT_DWH_DBNAME': str(conn.schema),
            'DBT_DWH_SCHEMA': dbt_schema_name,
            'DBT_PROFILES_DIR': folder,
        }
    elif dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
        analytics_conn = BaseHook.get_connection(dwh_conn_id)
        analytics_conn_extra = analytics_conn.extra_dejson
        env = {
            'DBT_ACCOUNT': analytics_conn_extra.get(
                'account',
                analytics_conn.host,
            ),
            'DBT_USER': analytics_conn.login,
            'DBT_PASS': analytics_conn.password,
            'DBT_ROLE': analytics_conn_extra.get('role'),
            'DBT_DB': analytics_conn_extra.get('database'),
            'DBT_WH': analytics_conn_extra.get('warehouse'),
            'DBT_SCHEMA': dbt_schema_name,
            'DBT_PROFILES_DIR': folder,
        }
    else:
        raise ValueError('DWH type not implemented!')

    #with dag:
    snsr = SqlSensor(
        task_id='sense_dbt_conflict_avoided',
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        dag=dag,
    )

    dbt_seed = BashOperator(
        task_id='run_dbt_seed',
        bash_command=bash_command.format('seed'),
        env=env,
        dag=dag,
    )

    dbt_run = BashOperator(
        task_id='run_dbt',
        bash_command=bash_command.format('run' + flags),
        env=env,
        dag=dag,
    )

    dbt_test = BashOperator(
        task_id='test_dbt',
        bash_command=bash_command.format('test' + flags),
        env=env,
        dag=dag,
    )

    dbt_docs = BashOperator(
        task_id='create_dbt_docs',
        bash_command=bash_command.format('docs generate'),
        env=env,
        dag=dag,
    )

    snsr >> dbt_seed >> dbt_run >> dbt_test

    if analytics_reader:
        # This should not occur when using Snowflake
        read_rights = PostgresOperator(
            task_id='grant_access_to_read_users',
            sql=analytics_reader_sql,
            postgres_conn_id=dwh_conn_id,
            dag=dag,
        )
        dbt_test >> read_rights >> dbt_docs
    else:
        dbt_test >> dbt_docs


    #with dag_full_refresh:
    snsr = SqlSensor(
        task_id='sense_dbt_conflict_avoided',
        conn_id=airflow_conn_id,
        sql=sensor_sql,
        dag=dag_full_refresh,
    )

    dbt_seed = BashOperator(
        task_id='run_dbt_seed',
        bash_command=bash_command.format('seed'),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_run = BashOperator(
        task_id='run_dbt',
        bash_command=bash_command.format('run --full-refresh' + flags),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_test = BashOperator(
        task_id='test_dbt',
        bash_command=bash_command.format('test' + flags),
        env=env,
        dag=dag_full_refresh,
    )

    dbt_docs = BashOperator(
        task_id='create_dbt_docs',
        bash_command=bash_command.format('docs generate'),
        env=env,
        dag=dag_full_refresh,
    )

    snsr >> dbt_seed >> dbt_run >> dbt_test

    if analytics_reader:
        read_rights = PostgresOperator(
            task_id='grant_access_to_read_users',
            sql=analytics_reader_sql,
            postgres_conn_id=dwh_conn_id,
            dag=dag_full_refresh,
        )
        dbt_test >> read_rights >> dbt_docs
    else:
        dbt_test >> dbt_docs
    return (dag, dag_full_refresh)



from airflow import DAG
from airflow.hooks.base_hook import BaseHook
from airflow.operators.postgres_operator import PostgresOperator as PGO
from airflow.sensors.external_task_sensor import ExternalTaskSensor as ETS

from ewah.ewah_utils.airflow_utils import etl_schema_tasks
from ewah.dwhooks.dwhook_snowflake import SnowflakeOperator
from ewah.constants import EWAHConstants as EC

from datetime import datetime, timedelta
from collections.abc import Iterable
from copy import deepcopy
import time

class ExtendedETS(ETS):
    """Extend ETS functionality to support the interplay of backfill and
    incremental DAGs."""

    def __init__(self,
        backfill_dag_id=None,
        backfill_execution_delta=None,
        backfill_execution_date_fn=None,
        backfill_external_task_id=None,
        execution_delay_in_seconds=120,
    *args, **kwargs):

        self.backfill_dag_id = backfill_dag_id
        self.backfill_execution_delta = backfill_execution_delta
        self.backfill_execution_date_fn = backfill_execution_date_fn
        self.backfill_external_task_id = backfill_external_task_id
        self.execution_delay_in_seconds = execution_delay_in_seconds

        super().__init__(*args, **kwargs)


    def execute(self, context):

        # When a DAG is executed as soon as possible, some data sources
        # may not immediately have up to date data from their API.
        # E.g. querying all data until 12.30pm only gives all relevant data
        # after 12.32pm due to some internal delays. In those cases, make
        # sure the incremental loading DAGs don't execute too quickly.
        next_execution_date = context['next_execution_date'] # type: Pendulum
        next_execution_date +=timedelta(seconds=self.execution_delay_in_seconds)
        while datetime.now() < next_execution_date:
            self.log.info('Waiting until {0} to execute... (now: {1})'.format(
                str(next_execution_date),
                str(datetime.now())
            ))
            time.sleep((self.execution_delay_in_seconds or 20) / 20)

        if context['dag'].start_date == context['execution_date']:
            # First execution of the DAG.
            if self.backfill_dag_id:
                # Check if the latest backfill ran! --> then run normally
                self.execution_delta = self.backfill_execution_delta or \
                    self.execution_delta
                self.execution_date_fn = self.backfill_execution_date_fn or \
                    self.execution_date_fn
                self.external_task_id = self.backfill_external_task_id or \
                    self.external_task_id
                self.external_dag_id = self.backfill_dag_id
                self.log.info('First instance, looking for previous backfill!')
                super().execute(context)
            else:
                self.log.info('This is the first execution of the DAG. Thus, ' \
                    + 'the sensor automatically succeeds.')
        else:
            super().execute(context)

def dag_factory_incremental_loading(
        dag_base_name,
        dwh_engine,
        dwh_conn_id,
        airflow_conn_id,
        start_date,
        el_operator,
        operator_config,
        target_schema_name,
        target_schema_suffix='_next',
        target_database_name=None,
        default_args=None,
        schedule_interval_backfill=timedelta(days=1),
        schedule_interval_future=timedelta(hours=1),
        switch_absolute_date=None, # If provided, switch from backfill DAG to
        #   normal DAG at that point in time (datetime.datetime) - best practice
        #   would be to leave None is most cases
        switch_relative_timedelta=None, # if switch_absolute_date is None, when
        #   to move from backfill DAG to normal DAG? Defaults to minuts half
        #   of schedule_interval_future (which is recommended in most cases)
        end_date=None,
        read_right_users=None,
        additional_dag_args={},
        additional_task_args={},
    ):

    if not hasattr(el_operator, '_IS_INCREMENTAL'):
        raise Exception('Invalid operator supplied!')
    if not el_operator._IS_INCREMENTAL:
        raise Exception('Operator does not support incremental loading!')
    if not type(schedule_interval_future) == timedelta:
        raise Exception('Schedule intervals must be datetime.timedelta!')
    if not type(schedule_interval_backfill) == timedelta:
        raise Exception('Schedule intervals must be datetime.timedelta!')
    if schedule_interval_backfill < timedelta(days=1):
        raise Exception('Backfill schedule interval cannot be below 1 day!')
    if schedule_interval_backfill < schedule_interval_future:
        raise Exception('Backfill schedule interval must be larger than' \
            + ' regular schedule interval!')
    if not operator_config.get('tables'):
        raise Exception('Requires a "tables" dictionary in operator_config!')
    if not (switch_relative_timedelta is None \
        or type(switch_relative_timedelta) == timedelta):
        raise Exception('switch_relative_timedelta must be timedelta or None!')
    if not read_right_users is None:
        if type(read_right_users) == str:
            read_right_users = read_right_users.split(',')
        if not isinstance(read_right_users, Iterable):
            raise Exception('read_right_users must be an iterable or string!')


    if not switch_absolute_date:
        if switch_relative_timedelta is None:
            # Make switch halway between latest normal DAG run and the
            #   next_execution_date of the next-to-run backfill DAG
            #   --> no interruption of the system, airflow has time to register
            #   the change, the backfill DAG can run once unimpeded and the
            #   normal DAG can then resume as per normal. Note: in that case,
            #   keep both DAGs active!
            switch_relative_timedelta = -schedule_interval_future / 2

        current_time = datetime.now() - switch_relative_timedelta
        # How much time has passed in total between start_date and now?
        switch_absolute_date = current_time - start_date
        # How often could the backfill DAG run in that time frame?
        switch_absolute_date /= schedule_interval_backfill
        switch_absolute_date = int(switch_absolute_date)
        # What is the exact datetime after the last of those runs?
        switch_absolute_date *= schedule_interval_backfill
        switch_absolute_date += start_date
        # --> switch_absolute_date is always in the (recent) past, unless
        #   switch_relative_timedelta is negative

    # Make sure that the backfill and normal DAG start_date and
    #   schedule_interval calculations were successful and correct
    backfill_timedelta = switch_absolute_date - start_date
    backfill_tasks_count = backfill_timedelta / schedule_interval_backfill
    # The schedule interval of the backfill must be an exact integer multiple
    # of the time period between start date and switch date!
    if not (backfill_tasks_count == round(backfill_tasks_count, 0)):
        raise Exception('The schedule interval of the backfill must be an ' \
            + 'exact integer multiple of the time period between start date '\
            + 'and switch date!')
    if end_date:
        backfill_end_date = min(switch_absolute_date, end_date)
    else:
        backfill_end_date = switch_absolute_date
    dags = (
        DAG(
            dag_base_name+'_Incremental',
            start_date=switch_absolute_date,
            end_date=end_date,
            schedule_interval=schedule_interval_future,
            catchup=True,
            max_active_runs=1,
            default_args=default_args,
            **additional_dag_args
        ),
        DAG(
            dag_base_name+'_Incremental_Backfill',
            start_date=start_date,
            end_date=backfill_end_date,
            schedule_interval=schedule_interval_backfill,
            catchup=True,
            max_active_runs=1,
            default_args=default_args,
            **additional_dag_args
        ),
        DAG(
            dag_base_name+'_Incremental_Reset',
            start_date=start_date,
            end_date=end_date,
            schedule_interval=None,
            catchup=False,
            max_active_runs=1,
            default_args=default_args,
            **additional_dag_args
        ),
    )

    # Create reset DAG
    reset_sql = """
        /*
            Different versions of airflow contain different tables. Only
            DELETE DAG from tables that actually exist.
        */
        CREATE OR REPLACE FUNCTION __ewah_delete_all_dag_stats(dag_name text)
        RETURNS void AS
        $$
        DECLARE
        	meta_db text;
        	meta_schema text;
        	meta_table text;
        BEGIN
        	FOR meta_db, meta_schema, meta_table IN
        		SELECT table_catalog, table_schema, table_name
        		FROM information_schema.tables
        		WHERE table_name IN ('dag_run', 'job', 'task_fail',
                    'task_instance', 'task_reschedule', 'xcom', 'dag_stats')
        		AND table_catalog LIKE %(db_name)s
        		AND table_schema LIKE %(schema_name)s
        	LOOP
        		EXECUTE 'DELETE FROM "' || meta_db || '"."' || meta_schema
                    || '"."' || meta_table
                    || '" WHERE dag_id = ''' || dag_name || '''';
        	END LOOP;
        END;
        $$ LANGUAGE plpgsql VOLATILE;
        SELECT __ewah_delete_all_dag_stats(%(dag_name)s);
        SELECT __ewah_delete_all_dag_stats(%(dag_name_backfill)s);
        DROP FUNCTION __ewah_delete_all_dag_stats;
    """
    airflow_conn = BaseHook.get_connection(airflow_conn_id)
    reset_task = PGO(
        sql=reset_sql,
        postgres_conn_id=airflow_conn_id,
        parameters={
            'dag_name': dag_base_name+'_Incremental',
            'dag_name_backfill': dag_base_name+'_Incremental_Backfill',
            'db_name': airflow_conn.schema,
            'schema_name': airflow_conn.extra_dejson.get('schema', 'public'),
        },
        task_id='reset_by_deleting_all_task_instances',
        dag=dags[2],
        **additional_task_args
    )
    drop_sql = f'DROP SCHEMA IF EXISTS "{target_schema_name}" CASCADE;'
    drop_sql += '\nDROP SCHEMA IF EXISTS "{schema}" CASCADE;'.format(**{
        'schema': target_schema_name + target_schema_suffix,
    })
    if dwh_engine == EC.DWH_ENGINE_POSTGRES:
        drop_task = PGO(
            sql=drop_sql,
            postgres_conn_id=dwh_conn_id,
            task_id='delete_previous_schema_if_exists',
            dag=dags[2],
            **additional_task_args
        )
    elif dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
        drop_task = SnowflakeOperator(
            sql=drop_sql,
            snowflake_conn_id=dwh_conn_id,
            database=target_database_name,
            task_id='delete_previous_schema_if_exists',
            dag=dags[2],
            **additional_task_args
        )
    else:
        raise ValueError('DWH not implemented for this task!')
    reset_task >> drop_task

    # Incremental DAG schema tasks
    kickoff, final = etl_schema_tasks(
        dag=dags[0],
        dwh_engine=dwh_engine,
        copy_schema=True,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        dwh_conn_id=dwh_conn_id,
        read_right_users=read_right_users,
        **additional_task_args
    )

    # Backfill DAG schema tasks
    kickoff_backfill, final_backfill = etl_schema_tasks(
        dag=dags[1],
        dwh_engine=dwh_engine,
        copy_schema=True,
        target_schema_name=target_schema_name,
        target_schema_suffix=target_schema_suffix,
        dwh_conn_id=dwh_conn_id,
        read_right_users=read_right_users,
        **additional_task_args
    )

    # Make sure incremental loading stops if there is an error!
    ets = (
        ExtendedETS(
            task_id='sense_previous_instance',
            allowed_states=['success', 'skipped'],
            external_dag_id=dags[0]._dag_id,
            external_task_id=final.task_id,
            execution_delta=schedule_interval_future,
            backfill_dag_id=dags[1]._dag_id,
            backfill_external_task_id=final_backfill.task_id,
            backfill_execution_delta=schedule_interval_backfill,
            dag=dags[0],
            **additional_task_args
        ),
        ExtendedETS(
            task_id='sense_previous_instance',
            allowed_states=['success', 'skipped'],
            external_dag_id=dags[1]._dag_id,
            external_task_id=final_backfill.task_id,
            execution_delta=schedule_interval_backfill,
            dag=dags[1],
            **additional_task_args
        )
    )
    ets[0] >> kickoff
    ets[1] >> kickoff_backfill

    # add table creation tasks
    count_backfill_tasks = 0
    for table in operator_config['tables'].keys():
        arg_dict = deepcopy(additional_task_args)
        arg_dict.update({'drop_and_replace': False})
        arg_dict.update(operator_config.get('general_config', {}))
        arg_dict_internal = {
            'task_id': 'extract_load_' + table,
            'dwh_engine': dwh_engine,
            'dwh_conn_id': dwh_conn_id,
            'target_table_name': table,
            'target_schema_name': target_schema_name,
            'target_schema_suffix': target_schema_suffix,
            'target_database_name': target_database_name,
            # 'drop_and_replace': False,
            # columns_definition
            # update_on_columns
            # primary_key_column_name
        }

        arg_dict_backfill = deepcopy(arg_dict)
        arg_dict.update(operator_config.get('incremental_config', {}))
        arg_dict_backfill.update(operator_config.get('backfill_config', {}))
        arg_dict.update(operator_config['tables'][table] or {})
        arg_dict_backfill.update(operator_config['tables'][table] or {})

        arg_dict.update(arg_dict_internal)
        arg_dict_backfill.update(arg_dict_internal)

        if not arg_dict.get('drop_and_replace', False):
            task_backfill = el_operator(dag=dags[1], **arg_dict_backfill)
            kickoff_backfill >> task_backfill >> final_backfill
            count_backfill_tasks += 1

        task = el_operator(dag=dags[0], **arg_dict)
        kickoff >> task >> final

    if count_backfill_tasks == 0:
        kickoff_backfill >> final_backfill

    return dags



from datetime import datetime
from yahoofinancials import YahooFinancials

from ewah.constants import EWAHConstants as EC
from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments

class EWAHFXOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    def __init__(
        self,
        currency_pair, # iterable of length 2
        data_from, # data_from is required
        data_until=None, # data_until defaults to current date
        frequency='daily', # daily, weekly, or monthly
    *args, **kwargs):

        if not type(data_from) in (str, datetime, type(None)):
            raise Exception('If supplied, data_from must be string, datetime' \
                + ', or None!')
        if not type(data_until) in (str, datetime, type(None)):
            raise Exception('If supplied, data_until must be string, datetime' \
                + ', or None!')

        if not frequency in ('daily', 'weekly', 'monthly'):
            raise Exception('Frequency must be one of: daily, weekly, monthly')

        if not len(currency_pair) == 2:
            raise Exception('currency_pair must be iterable of length 2 ' \
                + 'containing the currency pair.')

        # No connection id required for this operator
        if kwargs.get('source_conn_id'):
            self.log.info('source_conn_id is not required for operator! ' \
                + 'Ignoring argument.')
        kwargs['source_conn_id'] = None
        # If incremental, the primary key column is 'date'
        if kwargs.get('update_on_columns'):
            self.log.info('update_on_columns is fixed for this operator. ' \
                + 'Using the default.')
        kwargs['update_on_columns'] = ['date']
        super().__init__(*args, **kwargs)

        self.currency_pair = currency_pair
        self.data_from = data_from
        self.data_until = data_until
        self.frequency = frequency

    def execute(self, context):
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)

        format_str = '%Y-%m-%d'
        currency_str = '{0}{1}=X'.format(*self.currency_pair)
        data = YahooFinancials([currency_str]).get_historical_price_data(
            self.data_from.strftime(format_str),
            (self.data_until or datetime.now()).strftime(format_str),
            self.frequency,
        )
        self.upload_data(data[currency_str]['prices'])



from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from datetime import timedelta
from pytz import timezone

class EWAHSQLBaseOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until', 'reload_data_from')

    # implemented SQL sources - set self.sql_engine to this value in operator
    _MYSQL = 'MySQL'
    _PGSQL = 'PostgreSQL'
    _ORACLE = 'OracleSQL'

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = True

    def __init__(self,
        source_schema_name=None, # string
        source_table_name=None, # string, defaults to same as target_table_name
        data_from=None, # datetime, ISO datetime string, or airflow JINJA macro
        data_until=None, # datetime, ISO datetime string, or airflow JINJA macro
        timestamp_column=None, # name of column to increment and/or chunk by
        chunking_interval=None, # can be datetime.timedelta or integer
        chunking_column=None, # defaults to primary key if integer
        # also potentially used: primary_key_column_name of parent operator
        reload_data_from=None, # If a new table is added in production, and
        #   it is loading incrementally, where to start loading data? datetime
        reload_data_chunking=None, # must be timedelta
        where_clause=None,
    *args, **kwargs):

        source_table_name = source_table_name or kwargs.get('target_table_name')

        if not hasattr(self, 'sql_engine'):
            raise Exception('Operator invalid: need attribute sql_engine!')

        if reload_data_from and not (reload_data_chunking or chunking_interval):
            raise Exception('When setting reload_data_from, must also set ' \
                + 'either reload_data_chunking or chunking_interval!')

        if data_from or data_until:
            if not timestamp_column:
                raise Exception("If you used data_from and/or data_until, you" \
                    + " must also use timestamp_column to specify the column" \
                    + " that is being used!")

        if chunking_interval:
            if type(chunking_interval) == timedelta:
                if not timestamp_column:
                    raise Exception("If chunking via timedelta, must supply" \
                        + " a timestamp_column, even if not loading " \
                        + "incrementally!")
            elif type(chunking_interval) == int:
                if not chunking_column \
                    and not kwargs.get('primary_key_column_name'):
                    # Check columns for primary key - if exactly one,
                    # use it. Otherwise, raise error.
                    error_msg = "If chunking via integer, must supply " \
                        + "primary_key_column_name OR have EXACTLY ONE " \
                        + "primary key defined in the columns_definition dict" \
                        + "! This is not the case for: {0}. {{0}}" \
                            .format(source_table_name)
                    if not kwargs.get('columns_definition'):
                        raise Exception(error_msg.format(
                            'You did not supply the columns_definition dict.'
                        ))
                    columns = kwargs.get('columns_definition')
                    if not (sum([
                            1 if columns[key].get(EC.QBC_FIELD_PK) else 0
                            for key in columns.keys()
                        ]) == 1):
                        raise Exception(error_msg.format(
                            'There is not exactly one primary key in the dict.'
                        ))
                    for key in columns.keys():
                        if columns[key].get(EC.QBC_FIELD_PK):
                            kwargs['primary_key_column_name'] = key
                            break
                self.chunking_column = chunking_column \
                                        or kwargs['primary_key_column_name']

            else:
                raise Exception("Arg chunking_interval must be integer or "\
                    + "datetime.timedelta!")

        super().__init__(*args, **kwargs)

        if not self.drop_and_replace:
            if not (data_from and data_until):
                raise Exception('For incremental loading, you must specify ' \
                    + 'both data_from and data_until!')


        self.source_schema_name = source_schema_name
        self.source_table_name = source_table_name
        self.data_from = data_from
        self.data_until = data_until
        self.timestamp_column = timestamp_column
        self.chunking_interval = chunking_interval
        self.reload_data_from = reload_data_from
        self.reload_data_chunking = reload_data_chunking or chunking_interval
        self.where_clause = where_clause or '1 = 1'

    def execute(self, context):
        str_format = '%Y-%m-%dT%H:%M:%SZ'

        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.data_until = airflow_datetime_adjustments(self.data_until)
        self.reload_data_from = \
            airflow_datetime_adjustments(self.reload_data_from)

        if self.drop_and_replace:
            self.log.info('Loading data as full refresh.')
        else:
            if not self.test_if_target_table_exists():
                self.chunking_interval = self.reload_data_chunking \
                    or self.chunking_interval \
                    or (self.data_from - self.data_until)
                self.data_from = self.reload_data_from \
                    or context['dag'].start_date
                if type(self.data_from) == str:
                    self.data_from = datetime_from_string(self.data_from)

            self.log.info('Incrementally loading data from {0} to {1}.'.format(
                self.data_from.strftime(str_format),
                self.data_until.strftime(str_format),
            ))


        if self.columns_definition:
            sql_base = self._SQL_BASE_COLUMNS.format(**{
                'columns': ('{0}\n, {0}'.format(self._SQL_COLUMN_QUOTE)
                    .join(self.columns_definition.keys())),
                'schema': self.source_schema_name,
                'table': self.source_table_name,
            })
        else:
            sql_base = self._SQL_BASE_ALL.format(**{
                'schema': self.source_schema_name,
                'table': self.source_table_name,
            })

        params = {}
        # _SQL_PARAMS
        if self.drop_and_replace:
            if self.data_from:
                sql_base = sql_base.format('{0} >= {1} AND {{0}}'.format(
                    self.timestamp_column,
                    self._SQL_PARAMS.format('data_from'),
                ))
                params.update({'data_from': self.data_from})
            if self.data_until:
                sql_base = sql_base.format('{0} <= {1} AND {{0}}'.format(
                    self.timestamp_column,
                    self._SQL_PARAMS.format('data_until'),
                ))
                params.update({'data_until': self.data_from})
            sql_base = sql_base.format('{0} {{0}}'.format(self.where_clause))
        else:
            sql_base = sql_base.format('{0} >= {1} AND {0} < {2} AND {3} {{0}}'
                .format(
                    self.timestamp_column,
                    self._SQL_PARAMS.format('data_from'),
                    self._SQL_PARAMS.format('data_until'),
                    self.where_clause,
            ))
            params.update({'data_from': self.data_from})
            params.update({'data_until': self.data_until})


        if self.chunking_interval:
            if type(self.chunking_interval) == timedelta:
                chunking_column = self.timestamp_column
            else:
                chunking_column = self.chunking_column

            if self.drop_and_replace:
                previous_chunk, max_chunk = self._get_data_from_sql(
                    sql=self._SQL_MINMAX_CHUNKS.format(**{
                        'column': chunking_column,
                        'schema': self.source_schema_name,
                        'table': self.source_table_name,
                        'where_clause': self.where_clause,
                    }),
                    return_dict=False,
                )[0]
                if chunking_column == self.timestamp_column:
                    tz = timezone('UTC')
                    if not previous_chunk.tzinfo:
                        previous_chunk = tz.localize(previous_chunk)
                    if self.data_from:
                        previous_chunk = max(previous_chunk, self.data_from)
                    if not max_chunk.tzinfo:
                        max_chunk = tz.localize(max_chunk)
                    if self.data_until:
                        max_chunk = min(max_chunk, self.data_until)
            else:
                previous_chunk = self.data_from
                max_chunk = self.data_until

            while previous_chunk <= max_chunk:
                params.update({
                    'from_value': previous_chunk,
                    'until_value': min(
                        max_chunk,
                        previous_chunk + self.chunking_interval,
                    ),
                })
                data = self._get_data_from_sql(
                    sql=sql_base.format(
                        self._SQL_CHUNKING_CLAUSE
                    ).format(**{
                        'column': chunking_column,
                        'equal_sign': ('=' if max_chunk < (previous_chunk \
                            + self.chunking_interval) else ''),
                    }),
                    params=params,
                    return_dict=True,
                )

                self.upload_data(data=data)
                previous_chunk += self.chunking_interval
        else:
            self.upload_data(
                data=self._get_data_from_sql(
                    sql=sql_base.format('AND 1 = 1'),
                    return_dict=True,
                    params=params,
                ),
            )



"""
Modified version of the google analytics to s3 operator
Source of original: https://github.com/airflow-plugins/google_analytics_plugin/blob/master/operators/google_analytics_reporting_to_s3_operator.py
Accessed 5 March 2019
"""
from ewah.operators.base_operator import EWAHBaseOperator
from ewah.ewah_utils.airflow_utils import airflow_datetime_adjustments
from ewah.constants import EWAHConstants as EC

from airflow.hooks.base_hook import BaseHook

import json
import time
from datetime import timedelta
from apiclient.discovery import build
from oauth2client.service_account import ServiceAccountCredentials as SAC


class EWAHGAOperator(EWAHBaseOperator):

    template_fields = ('data_from', 'data_until')

    _IS_INCREMENTAL = True
    _IS_FULL_REFRESH = False

    _API_CORE_V3 = 'core_v3'
    _API_CORE_V4 = 'core_v4'
    _API_MULTI = 'multichannel'

    _ACCEPTED_API = [
        # _API_CORE_V3,
        _API_CORE_V4,
        # _API_MULTI,
    ]

    def __init__(
        self,
        api, # one of _API_CORE_V3, _API_CORE_V4, _API_MULTI
        view_id,
        data_from,
        data_until,
        dimensions,
        metrics,
        page_size=10000,
        include_empty_rows=True,
        sampling_level=None,
        chunking_interval=None, # must be timedelta
        reload_data_from=None, # If a new table is added in production, and
        #   it is loading incrementally, where to start loading data? datetime
        reload_data_chunking=None, # must be timedelta
        # Note: Need chunking despite pagination as DWH upload must also be
        #   chunked, not just data loading, especially for reloads / backfills!
    *args, **kwargs):
        if kwargs.get('update_on_columns'):
            raise Exception('update_on_columns supplied, but the field is ' \
                + 'auto-generated by the operator!')
        if reload_data_from and not (reload_data_chunking or chunking_interval):
            raise Exception('When setting reload_data_from, must also set ' \
                + 'either reload_data_chunking or chunking_interval!')

        if not api in self._ACCEPTED_API:
            raise Exception('api must be one of these: {0}'.format(
                ', '.join(self._ACCEPTED_API),
            ))

        if api == self._API_MULTI:
            shorthand = 'mcf:'
        else:
            shorthand = 'ga:'

        dimensions = [
            ('' if dim.startswith(shorthand) else shorthand) + dim
            for dim in dimensions
        ]
        metrics = [
            ('' if metric.startswith(shorthand) else shorthand) + metric
            for metric in metrics
        ]

        kwargs.update({'update_on_columns': [dim[3:] for dim in dimensions]})

        super().__init__(*args, **kwargs)

        self.credentials = BaseHook.get_connection(self.source_conn_id)
        self.credentials = self.credentials.extra_dejson
        self.api = api
        self.view_id = view_id
        self.data_from = data_from
        self.data_until = data_until
        self.sampling_level = sampling_level
        self.dimensions = dimensions
        self.metrics = metrics
        self.page_size = page_size
        self.include_empty_rows = include_empty_rows
        self.chunking_interval = chunking_interval
        self.reload_data_from = reload_data_from
        self.reload_data_chunking = reload_data_chunking or chunking_interval

        self.metricMap = {
            'METRIC_TYPE_UNSPECIFIED': 'varchar(255)',
            'CURRENCY': 'decimal(20,5)',
            'INTEGER': 'int(11)',
            'FLOAT': 'decimal(20,5)',
            'PERCENT': 'decimal(20,5)',
            'TIME': 'time'
        }

        if chunking_interval and not (type(chunking_interval) == timedelta):
            raise Exception('If supplied, chunking_interval must be timedelta!')


        if not self.credentials.get('client_secrets'):
            raise Exception('Google Analytics Credentials misspecified!' \
                + ' Example of a correct specifidation: {0}'.format(
                    json.dumps({"client_secrets":{
                        "type": "service_account",
                        "project_id": "abc-123",
                        "private_key_id": "123456abcder",
                        "private_key": "-----BEGIN PRIVATE KEY-----\nxxx\n-----END PRIVATE KEY-----\n",
                        "client_email": "xyz@abc-123.iam.gserviceaccount.com",
                        "client_id": "123457",
                        "auth_uri": "https://accounts.google.com/o/oauth2/auth",
                        "token_uri": "https://oauth2.googleapis.com/token",
                        "auth_provider_x509_cert_url": "https://www.googleapis.com/oauth2/v1/certs",
                        "client_x509_cert_url": "https://www.googleapis.com/robot/v1/metadata/x509/xyz%40abc-123.iam.gserviceaccount.com"
                    }})
                ))

        if len(dimensions) > 7:
            raise Exception(('Can only fetch up to 7 dimensions!' \
                + ' Currently {0} Dimensions').format(
                    str(len(dimensions)),
            ))

        if len(metrics) > 10:
            raise Exception(('Can only fetch up to 10 metrics!' \
                + ' Currently {0} Dimensions').format(
                    str(len(metrics)),
            ))

        if self.page_size > 10000:
            raise Exception(
                'Please specify a page size equal to or lower than 10000.')


    def execute(self, context):
        self.data_until = airflow_datetime_adjustments(self.data_until)
        self.data_from = airflow_datetime_adjustments(self.data_from)
        self.reload_data_from = \
            airflow_datetime_adjustments(self.reload_data_from)

        if self.drop_and_replace or (not self.test_if_target_table_exists()):
            self.chunking_interval = self.reload_data_chunking \
                or self.chunking_interval \
                or (self.data_until - self.data_from)
            self.data_from = self.reload_data_from or context['dag'].start_date
            self.data_from = airflow_datetime_adjustments(self.data_from)
        elif not self.chunking_interval:
            self.chunking_interval = self.data_until - self.data_from

        self.log.info('Connecting to Google...')
        if self.api == self._API_CORE_V3:
            raise Exception('Not yet implemented')
        elif self.api == self._API_CORE_V4:
            kwargs = {
                'service_object': build(
                    'analyticsreporting',
                    'v4',
                    credentials=SAC.from_json_keyfile_dict(
                        self.credentials['client_secrets'],
                        ['https://www.googleapis.com/auth/analytics.readonly'],
                    )
                ),
                'context': context,
            }
            data_func = self.get_data_v4
        elif self.api == self._API_MULTI:
            raise Exception('Not yet implemented!')
            # Open: how to parse raw data to uploadable format? See get_data_multi
            kwargs = {
                'service_object': build(
                    'analytics',
                    'v3',
                    credentials=SAC.from_json_keyfile_dict(
                        self.credentials['client_secrets'],
                        ['https://www.googleapis.com/auth/analytics.readonly'],
                    )
                ),
                'context': context,
            }
            data_func = self.get_data_multi
        else:
            raise Exception('Not yet implemented!')

        i = 0
        while self.data_from < self.data_until:
            i += 1

            self.since_formatted = self.data_from.strftime('%Y-%m-%d')
            self.until_formatted = min(
                    self.data_from + self.chunking_interval,
                    self.data_until,
                ).strftime('%Y-%m-%d')
            self.log.info('Chunk {0}: From {1} through {2}'.format(
                str(i),
                self.since_formatted,
                self.until_formatted,
            ))

            self.upload_data(data_func(**kwargs))
            self.data_from += self.chunking_interval

    def get_data_multi(self, service_object, context):
        start_index = 1
        payload = {
            'ids': 'ga:' + self.view_id,
            'start_date': self.since_formatted,
            'end_date': self.until_formatted,
            'metrics': ','.join(self.metrics),
            'dimensions': ','.join(self.dimensions),
            'sort': ','.join(self.dimensions),
            # 'filters': not implemented!
            # 'samplingLevel': Default!
            'start_index': start_index,
            'max_results': self.page_size,
        }

        self.log.info('Getting data from page 1...')
        i = 1
        response = service_object.data().mcf().get(**payload).execute()
        data = response.get('rows', [])
        while response.get('nextLink'):
            i += 1
            self.log.info('Getting data from page {0}...'.format(str(i)))
            start_index += self.page_size
            payload.update({'start_index': start_index})
            response = service_object.data().mcf().get(**payload).execute()
            data += response,get('rows', [])

        uploadable_data = [] # TBU

        return uploadable_data

    def get_data_v4(self, service_object, context):
        reportRequest = {
            'viewId': self.view_id,
            'dateRanges': [{
                'startDate': self.since_formatted,
                'endDate': self.until_formatted,
            }],
            'samplingLevel': self.sampling_level,
            'dimensions': [{'name':d} for d in self.dimensions],
            'metrics': [{'expression':m} for m in self.metrics],
            'pageSize': self.page_size,
            'includeEmptyRows': self.include_empty_rows,
        }

        response = (service_object
                    .reports()
                    .batchGet(body={'reportRequests':[reportRequest]})
                    .execute())

        if response.get('reports'):
            report = response['reports'][0]
            rows = report.get('data', {}).get('rows', [])

            while report.get('nextPageToken'):
                time.sleep(1)
                reportRequest.update({'pageToken': report['nextPageToken']})
                response = (service_object
                            .reports()
                            .batchGet(body={'reportRequests':[reportRequest]})
                            .execute())
                report = response['reports'][0]
                rows.extend(report.get('data', {}).get('rows', []))

            if report['data']:
                report['data']['rows'] = rows
        else:
            report = {}

        columnHeader = report.get('columnHeader', {})
        # Right now all dimensions are hardcoded to varchar(255), will need a
        # map if any non-varchar dimensions are used in the future
        # Unfortunately the API does not send back types for Dimensions like it
        # does for Metrics (yet..)
        dimensionHeaders = [
            {
                'name': header.replace('ga:', ''),
                'type': 'varchar(255)',
            }
            for header
            in columnHeader.get('dimensions', [])
        ]
        metricHeaders = [
            {
                'name': entry.get('name').replace('ga:', ''),
                'type': self.metricMap.get(entry.get('type'), 'varchar(255)'),
            }
            for entry in
            columnHeader.get('metricHeader', {}).get('metricHeaderEntries', [])
        ]

        uploadable_data = []
        rows = report.get('data', {}).get('rows', [])
        for row_counter, row in enumerate(rows):
            root_data_obj = {}
            dimensions = row.get('dimensions', [])
            metrics = row.get('metrics', [])

            for index, dimension in enumerate(dimensions):
                header = dimensionHeaders[index].get('name') # .lower()
                root_data_obj[header] = dimension

            for metric in metrics:
                data = {}
                data.update(root_data_obj)

                for index, value in enumerate(metric.get('values', [])):
                    header = metricHeaders[index].get('name') # .lower()
                    data[header] = value

                data['view_id'] = self.view_id

                uploadable_data += [data]

        return uploadable_data



from airflow.operators.postgres_operator import PostgresOperator as PGO
from airflow.operators.python_operator import PythonOperator as PO
from airflow.hooks.base_hook import BaseHook

from ewah.dwhooks.dwhook_snowflake import EWAHDWHookSnowflake
from ewah.constants import EWAHConstants as EC

from datetime import datetime, timedelta, timezone
from copy import deepcopy
import re

def airflow_datetime_adjustments(datetime_raw):
    if type(datetime_raw) == str:
        datetime_string = datetime_raw
        if '-' in datetime_string[10:]:
            tz_sign = '-'
        else:
            tz_sign = '+'
        datetime_strings = datetime_string.split(tz_sign)

        if 'T' in datetime_string:
            format_string = '%Y-%m-%dT%H:%M:%S'
        else:
            format_string = '%Y-%m-%d %H:%M:%S'

        if '.' in datetime_string:
            format_string += '.%f'

        if len(datetime_strings) == 2:
            if ':' in datetime_strings[1]:
                datetime_strings[1] = datetime_strings[1].replace(':', '')
            datetime_string = tz_sign.join(datetime_strings)
            format_string += '%z'
        elif 'Z' in datetime_string:
            format_string += 'Z'
        datetime_raw = datetime.strptime(
            datetime_string,
            format_string,
        )
    elif not (type(datetime_raw) in (type(None), datetime)):
        raise Exception('Invalid datetime type supplied! Supply either string' \
            + ' or datetime.datetime! supplied: {0}'.format(
                str(type(datetime_raw))
            ))

    if datetime_raw and datetime_raw.tzinfo is None:
        datetime_raw = datetime_raw.replace(tzinfo=timezone.utc)

    return datetime_raw

def etl_schema_tasks(
        dag,
        dwh_engine,
        dwh_conn_id,
        target_schema_name,
        target_schema_suffix='_next',
        target_database_name=None,
        copy_schema=False,
        read_right_users=None, # Only for PostgreSQL
        **additional_task_args
    ):

    if dwh_engine == EC.DWH_ENGINE_POSTGRES:
        if copy_schema:
            sql_kickoff = '''
                /*
                 * Clone schema function found here: https://wiki.postgresql.org/wiki/Clone_schema
                 * Accessed 9 March 2019
                 * The function is created/replaced in this query to make sure it is
                 * available when called thereafter.
                 * Added an INSERT statement to copy data into the new schema as well.
                 */
                CREATE SCHEMA IF NOT EXISTS "{schema_name}" /* for edge case where schema is created for the first time! */;
                CREATE OR REPLACE FUNCTION "{schema_name}".clone_schema(source_schema text, dest_schema text) RETURNS void AS
                $$

                DECLARE
                  object text;
                  buffer text;
                  default_ text;
                  column_ text;
                BEGIN
                  EXECUTE 'CREATE SCHEMA "' || dest_schema || '"';

                  -- TODO: Find a way to make this sequence's owner is the correct table.
                  FOR object IN
                    SELECT sequence_name::text FROM information_schema.SEQUENCES WHERE sequence_schema = source_schema
                  LOOP
                    EXECUTE 'CREATE SEQUENCE "' || dest_schema || '"."' || object || '"';
                  END LOOP;

                  FOR object IN
                    SELECT TABLE_NAME::text FROM information_schema.TABLES WHERE table_schema = source_schema
                  LOOP
                    buffer := '"' || dest_schema || '"."' || object || '"';
                    EXECUTE 'CREATE TABLE ' || buffer || ' (LIKE "' || source_schema || '"."' || object || '" INCLUDING CONSTRAINTS INCLUDING INDEXES INCLUDING DEFAULTS)';

                    FOR column_, default_ IN
                      SELECT column_name::text, REPLACE(column_default::text, source_schema, dest_schema) FROM information_schema.COLUMNS WHERE table_schema = dest_schema AND TABLE_NAME = object AND column_default LIKE 'nextval(%' || source_schema || '%::regclass)'
                    LOOP
                      EXECUTE 'ALTER TABLE ' || buffer || ' ALTER COLUMN "' || column_ || '" SET DEFAULT ' || default_;
                    END LOOP;
                    EXECUTE 'INSERT INTO ' || buffer || ' SELECT * FROM "' || source_schema || '"."' || object || '"';
                  END LOOP;

                END;

                $$ LANGUAGE plpgsql VOLATILE;

                /*
                 * Clone the schema
                 */
                DROP SCHEMA IF EXISTS "{schema_name}{schema_suffix}" CASCADE;
                SELECT "{schema_name}".clone_schema('{schema_name}', '{schema_name}{schema_suffix}');
                '''.format(**{
                    'schema_name': target_schema_name,
                    'schema_suffix': target_schema_suffix,
                })
        else:
            sql_kickoff = ''' /* Create a new, empty schema */
                DROP SCHEMA IF EXISTS "{schema_name}{schema_suffix}" CASCADE;
                CREATE SCHEMA "{schema_name}{schema_suffix}";
                '''.format(**{
                    'schema_name': target_schema_name,
                    'schema_suffix': target_schema_suffix,
                })
        sql_final = '''
            DROP SCHEMA IF EXISTS "{schema_name}" CASCADE;
            ALTER SCHEMA "{schema_name}{schema_suffix}" RENAME TO "{schema_name}";
            '''.format(**{
                'schema_name': target_schema_name,
                'schema_suffix': target_schema_suffix,
            })

        if read_right_users:
            if not type(read_right_users) == list:
                raise Exception('Arg read_right_users must be of type List!')
            for user in read_right_users:
                if re.search(r"\s", user) or (';' in user):
                    raise ValueError('No whitespace or semicolons allowed in usernames!')
                sql_final += f'\nGRANT USAGE ON SCHEMA "{target_schema_name}" TO {user};'
                sql_final += f'\nGRANT SELECT ON ALL TABLES IN SCHEMA "{target_schema_name}" TO {user};'

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update({
            'sql': sql_kickoff,
            'task_id': 'kickoff_{0}'.format(target_schema_name),
            'dag': dag,
            'postgres_conn_id': dwh_conn_id,
            # 'retries': 1,
            # 'retry_delay': timedelta(minutes=1),
        })
        task_2_args.update({
            'sql': sql_final,
            'task_id': 'final_{0}'.format(target_schema_name),
            'dag': dag,
            'postgres_conn_id': dwh_conn_id,
            # 'retries': 1,
            # 'retry_delay': timedelta(minutes=1),
        })
        return (PGO(**task_1_args), PGO(**task_2_args))
    elif dwh_engine == EC.DWH_ENGINE_SNOWFLAKE:
        target_database_name = target_database_name or (BaseHook \
                .get_connection(dwh_conn_id).extra_dejson.get('database'))
        if copy_schema:
            sql_kickoff = '''
                /*Make sure there is something to clone in step 2*/
                DROP SCHEMA IF EXISTS "{database}"."{schema_name}{schema_suffix}" CASCADE;
                CREATE SCHEMA IF NOT EXISTS "{database}"."{schema_name}";
                CREATE SCHEMA "{database}"."{schema_name}{schema_suffix}"
                CLONE "{database}"."{schema_name}";
                '''.format(**{
                    'database': target_database_name,
                    'schema_name': target_schema_name,
                    'schema_suffix': target_schema_suffix,
                })
        else:
            sql_kickoff = '''
                DROP SCHEMA IF EXISTS "{database}"."{schema_name}{schema_suffix}" CASCADE;
                CREATE SCHEMA "{database}"."{schema_name}{schema_suffix}";
                '''.format(**{
                    'database': target_database_name,
                    'schema_name': target_schema_name,
                    'schema_suffix': target_schema_suffix,
                })
        sql_final = '''
            DROP SCHEMA IF EXISTS "{database}"."{schema_name}" CASCADE;
            ALTER SCHEMA "{database}"."{schema_name}{schema_suffix}" RENAME TO "{schema_name}";
            '''.format(**{
                'database': target_database_name,
                'schema_name': target_schema_name,
                'schema_suffix': target_schema_suffix,
            })

        def execute_snowflake(sql, conn_id, **kwargs):
            hook = EWAHDWHookSnowflake(conn_id)
            hook.execute(sql)
            hook.close()

        task_1_args = deepcopy(additional_task_args)
        task_2_args = deepcopy(additional_task_args)
        task_1_args.update({
            'task_id': 'kickoff_{0}'.format(target_schema_name),
            'python_callable': execute_snowflake,
            'op_kwargs': {
                'sql': sql_kickoff,
                'conn_id': dwh_conn_id,
            },
            'provide_context': True,
            'dag': dag,
        })
        task_2_args.update({
            'task_id': 'final_{0}'.format(target_schema_name),
            'python_callable': execute_snowflake,
            'op_kwargs': {
                'sql': sql_final,
                'conn_id': dwh_conn_id,
            },
            'provide_context': True,
            'dag': dag,
        })
        return (PO(**task_1_args), PO(**task_2_args))
    else:
        raise ValueError('Feature not implemented!')



from airflow import DAG
from airflow.operators.bash_operator import BashOperator

def git_pull_dag_factory(
    dag_name,
    folder,
    default_args={},
    start_date=None,
    schedule_interval=None,
    ):

    dag = DAG(
        dag_name,
        catchup=False,
        max_active_runs=1,
        schedule_interval=schedule_interval,
        start_date=start_date,
        default_args=default_args,
    )

    task = BashOperator(
        task_id='git_pull_task',
        bash_command='cd {0} && git pull'.format(folder),
        dag=dag,
    )

    return dag



from ewah.operators.sql_base_operator import EWAHSQLBaseOperator
# Use EWAHDWHookPostgres as hook since it's already there - not necessarily SOP!
from ewah.dwhooks.dwhook_postgres import EWAHDWHookPostgres

class EWAHPostgresOperator(EWAHSQLBaseOperator):

    _SQL_BASE_COLUMNS = 'SELECT\n"{columns}"\nFROM "{schema}"."{table}"\nWHERE {{0}};'
    _SQL_BASE_ALL = 'SELECT * FROM "{schema}"."{table}"\nWHERE {{0}};'
    _SQL_COLUMN_QUOTE = '"'
    _SQL_MINMAX_CHUNKS = 'SELECT MIN({column}), MAX({column}) FROM "{schema}"."{table}" WHERE {where_clause};'
    _SQL_CHUNKING_CLAUSE = 'AND {column} >= %(from_value)s AND {column} <{equal_sign} %(until_value)s'
    _SQL_PARAMS = '%({0})s'

    def __init__(self, *args, **kwargs):
        self.sql_engine = self._PGSQL
        super().__init__(*args, **kwargs)

    def _get_data_from_sql(self, sql, params=None, return_dict=True):
        hook = EWAHDWHookPostgres(self.source_conn_id)
        self.log.info('Executing:\n{0}\n\nWith params:\n{1}'.format(
            sql,
            str(params),
        ))
        data = hook.execute_and_return_result(
            sql=sql,
            params=params,
            return_dict=return_dict,
        )
        hook.close()
        return data



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ewah", # Replace with your own username
    version="0.1.19",
    author="Bijan Soltani",
    author_email="bijan.soltani+ewah@gemmaanalytics.com",
    description="An ELT with airflow helper module: Ewah",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gemmaanalytics.com/",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
    install_requires=[            # I get to this in a second
          'pyyaml',
          'psycopg2',
          'gspread',
          'pytz',
          'yahoofinancials',
          'google-api-python-client',
          'oauth2client',
          'cx_Oracle',
          'facebook_business',
          'mysql-connector-python',
          'snowflake-connector-python',
          'pymongo',
      ],
)



# Adapted from https://github.com/teamclairvoyant/airflow-maintenance-dags
from airflow.models import DAG, Variable
from airflow.operators import BashOperator
from airflow.configuration import conf
from datetime import datetime, timedelta
import os

#from c_utils.various_constants import AirflowConstants
#acs = AirflowConstants()

"""
A maintenance workflow that you can deploy into Airflow to periodically clean out the task logs to avoid those getting too big.

airflow trigger_dag --conf '{"maxLogAgeInDays":30}' airflow-log-cleanup

--conf options:
    maxLogAgeInDays:<INT> - Optional

"""

def cleanup_dag_factory(
    dag_name='utils_airflow_log_cleanup',
    state_date=datetime(2018, 8, 23),
    schedule_interval=timedelta(days=1),
    max_log_age_in_days=7,
    dag_owner='utils',
    alert_emails=[],
):
    DAG_ID = dag_name # os.path.basename(__file__).replace(".pyc", "").replace(".py", "")  # airflow-log-cleanup
    START_DATE = state_date
    BASE_LOG_FOLDER = conf.get("core", "BASE_LOG_FOLDER")
    SCHEDULE_INTERVAL = schedule_interval
    DAG_OWNER_NAME = dag_owner       # Who is listed as the owner of this DAG in the Airflow Web Server
    ALERT_EMAIL_ADDRESSES = alert_emails          # List of email address to send email alerts to if this job fails
    DEFAULT_MAX_LOG_AGE_IN_DAYS = max_log_age_in_days # Length to retain the log files if not already provided in the conf. If this is set to 30, the job will remove those files that are 30 days old or odler
    ENABLE_DELETE = True                # Whether the job should delete the logs or not. Included if you want to temporarily avoid deleting the logs
    NUMBER_OF_WORKERS = 1               # The number of worker nodes you have in Airflow. Will attempt to run this process for however many workers there are so that each worker gets its logs cleared.

    default_args = {
        'email': ALERT_EMAIL_ADDRESSES,
        'email_on_failure': bool(ALERT_EMAIL_ADDRESSES),
        'owner': dag_owner,
        'email_on_retry': False,
        'start_date': START_DATE,
        'retries': 1,
        'retry_delay': timedelta(minutes=1)
    }

    dag = DAG(DAG_ID, default_args=default_args, schedule_interval=SCHEDULE_INTERVAL, start_date=START_DATE, catchup=False)

    log_cleanup = """
    echo "Getting Configurations..."
    BASE_LOG_FOLDER='""" + BASE_LOG_FOLDER + """'
    MAX_LOG_AGE_IN_DAYS="{{dag_run.conf.maxLogAgeInDays}}"
    if [ "${MAX_LOG_AGE_IN_DAYS}" == "" ]; then
        echo "maxLogAgeInDays conf variable isn't included. Using Default '""" + str(DEFAULT_MAX_LOG_AGE_IN_DAYS) + """'."
        MAX_LOG_AGE_IN_DAYS='""" + str(DEFAULT_MAX_LOG_AGE_IN_DAYS) + """'
    fi
    ENABLE_DELETE=""" + str("true" if ENABLE_DELETE else "false") + """
    echo "Finished Getting Configurations"
    echo ""

    echo "Configurations:"
    echo "BASE_LOG_FOLDER:      '${BASE_LOG_FOLDER}'"
    echo "MAX_LOG_AGE_IN_DAYS:  '${MAX_LOG_AGE_IN_DAYS}'"
    echo "ENABLE_DELETE:        '${ENABLE_DELETE}'"
    echo ""

    echo "Running Cleanup Process..."
    FIND_STATEMENT="find ${BASE_LOG_FOLDER}/*/* -type f -mtime +${MAX_LOG_AGE_IN_DAYS}"
    echo "Executing Find Statement: ${FIND_STATEMENT}"
    FILES_MARKED_FOR_DELETE=`eval ${FIND_STATEMENT}`
    echo "Process will be Deleting the following directories:"
    echo "${FILES_MARKED_FOR_DELETE}"
    echo "Process will be Deleting `echo "${FILES_MARKED_FOR_DELETE}" | grep -v '^$' | wc -l ` file(s)"     # "grep -v '^$'" - removes empty lines. "wc -l" - Counts the number of lines
    echo ""

    if [ "${ENABLE_DELETE}" == "true" ];
    then
        DELETE_STMT="${FIND_STATEMENT} -delete"
        echo "Executing Delete Statement: ${DELETE_STMT}"
        eval ${DELETE_STMT}
        DELETE_STMT_EXIT_CODE=$?
        if [ "${DELETE_STMT_EXIT_CODE}" != "0" ]; then
            echo "Delete process failed with exit code '${DELETE_STMT_EXIT_CODE}'"
            exit ${DELETE_STMT_EXIT_CODE}
        fi
    else
        echo "WARN: You're opted to skip deleting the files!!!"
    fi
    echo "Finished Running Cleanup Process"
    """

    for log_cleanup_id in range(1, NUMBER_OF_WORKERS + 1):

        log_cleanup = BashOperator(
            task_id='log_cleanup_' + str(log_cleanup_id),
            bash_command=log_cleanup,
            provide_context=True,
            dag=dag)

    return dag
